#!/usr/bin/env python
# coding: utf-8
#!/usr/bin/env bash
import pandas as pd
import numpy as np
from sklearn.cluster import KMeans,AffinityPropagation,AgglomerativeClustering,Birch,DBSCAN,MeanShift,OPTICS
from sklearn.cluster import SpectralClustering
from sklearn.mixture import GaussianMixture
from sklearn.metrics import silhouette_score, calinski_harabasz_score, davies_bouldin_score
import os,sys
sys.path.append(os.path.dirname(os.path.dirname(__file__)))
from FreeAeonML.FACommon import CFACommon
from FreeAeonML.FASample import CFASample

class CFAModelCluster:
    def __init__(self,cluster_count,models = None):
        if models == None:
            self.m_models = {
                "KMeans": KMeans(n_clusters=cluster_count, random_state=False),
                "AffinityPropagation":AffinityPropagation(damping=0.9),
                "AgglomerativeClustering": AgglomerativeClustering(n_clusters=cluster_count),
                "Birch": Birch(threshold=0.01, n_clusters=cluster_count),
                #"DBSCAN": DBSCAN(eps=0.30, min_samples=9),
                "MeanShift":MeanShift(),
                "OPTICS":OPTICS(eps=0.30, min_samples=9),
                "SpectralClustering": SpectralClustering(n_clusters=cluster_count, random_state=False),
                "GaussianMixture": GaussianMixture(n_components=cluster_count, covariance_type='full', random_state=False)
            }
        else:
            self.m_models = models

    def fit_predict(self,df_sample):
        X_train = df_sample.to_numpy()
        cluster_result = []
        for model_name in self.m_models:
            model = self.m_models[model_name]
            yhat = model.fit_predict(X_train)
            clusters = np.unique(yhat)
            for cluster in clusters:
                row_ix = np.where(yhat == cluster)
                tmp = {}
                tmp['model'] = model_name
                tmp['cluster'] = cluster
                tmp['rows'] = row_ix
                cluster_result.append(tmp)
        return pd.DataFrame(cluster_result)

    def evaluate(self,df_sample):
        X_train = df_sample.to_numpy()
        result = []
        for model_name in self.m_models.keys():
            print(model_name)
            model = self.m_models[model_name]
            tmp = {"model_name": model_name}
            yhat = model.fit_predict(X_train)
            tmp["silhouette_score"] = silhouette_score(X_train, yhat)
            tmp["calinski_harabasz_score"] = calinski_harabasz_score(X_train, yhat)
            tmp["davies_bouldin_score"] = davies_bouldin_score(X_train, yhat)
            result.append(tmp)
        return pd.DataFrame(result)

    def sample_cluster(self,df_cluster_result,df_sample,model_name):
        df_tmp = df_cluster_result[df_cluster_result['model'] == model_name]
        df_ret = df_sample.copy(deep = True)
        df_ret['_cluster'] = ""
        for rows, cluster in zip(df_tmp['rows'], df_tmp['cluster']):
            indexes = rows[0].astype(int)
            df_ret.iloc[indexes, -1] = cluster
        return df_ret

def main():
    df_sample = CFASample.get_random_cluster()
    test = CFAModelCluster(2)
    df_result = test.fit_predict(df_sample)
    print(df_result)
    df_perf = test.evaluate(df_sample)
    print(df_perf)
    df_cluster_sample = test.sample_cluster(df_result,df_sample,"KMeans")
    print(df_cluster_sample)
if __name__ == "__main__":
    main()



'''
数据预处理
1. CFADataProcess (数据预处理)
   --将数据转换为正态分布（box-cox）
   --从正态分布中，将数据还原（box-cox）
   --获取异常数据（Z-Score）
2. CFATransformer(数据转换)
   --将原始数据转换另一组数据风格
   --scale: 线性缩放（不改变分布）
   --quantile: 改变分布，分位数映射（支持单变量或多变量）
   --copula:   改变分布，多变量copula变换（默认）
'''
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from scipy import stats
from scipy.stats import norm
from scipy.special import inv_boxcox
from sklearn.preprocessing import StandardScaler
import os,sys
sys.path.append(os.path.dirname(os.path.dirname(__file__)))
from FreeAeonML.FASample import CFASample

class CFADataPreprocess:
    def __init__(self):
        pass
    '''
    将数据转换为正态分布（使用box-cox）
    lambda_value==0，使用log转换，lambda_value==None，自动计算最近lambda参数
    '''
    @staticmethod
    def normal_transform(ds_data,lambda_value=None):
        t_data, l_value = stats.boxcox(ds_data,lambda_value)
        return pd.Series(t_data,index=ds_data.index),l_value
    '''
    从正态分布中，将数据还原（使用box-cox）
    '''
    @staticmethod
    def normal_recovery(ds_data,lambda_value):
        original_data = inv_boxcox(np.array(ds_data.values),lambda_value)
        return pd.Series(original_data,index=ds_data.index)
    '''
    多项式拟合
    ds_x：pd.Series,自变量，当为空时，时间序列数据拟合。
    ds_y：pd.Series,因变量
    degree：多项式最高阶数
    返回值：拟合对象(p_object),拟合残差的标准差(e_std),拟合后的数据(v_object)
    '''
    def polyfit(ds_y,ds_x = pd.Series([]), degree = 2):
        if ds_x.empty :
            x = np.arange(ds_y.shape[0]) # 生成对应的时间点作为（自变量）
        else:
            x = ds_x.values()
        y = ds_y.values
    
        p_coefficients, e_residuals, _, _, _  = np.polyfit(x, y, degree,full=True)
    
        p_object = np.poly1d(p_coefficients)
        v_object = pd.Series(p_object(x),index=ds_y.index)
        e_std = np.sqrt(e_residuals / len(x))
        return p_object,e_std,v_object
    '''
    获取异常数据（Z-Score绝对值大于3 sigma）
    ''' 
    @staticmethod
    def get_abnormal(ds_data,n_sigma=3):
        ds_z_score = (ds_data - ds_data.mean()) / ds_data.std()
        filtered =  ds_z_score[(ds_z_score<(0-n_sigma) ) | (ds_z_score>n_sigma)]
        return ds_data.iloc[filtered.index]
    '''
    去掉异常数据（Z-Score绝对值大于3 sigma）
    ''' 
    @staticmethod
    def remove_abnormal(ds_data,n_sigma=3):
        ds_z_score = (ds_data - ds_data.mean()) / ds_data.std()
        filtered =  ds_z_score[(ds_z_score>(0-n_sigma) ) & (ds_z_score < n_sigma)]
        return ds_data.iloc[filtered.index]
        
    # 对特征字段进行标准化处理(z-score/max-min)
    @staticmethod
    def get_scale(df_data,x_columns=[],y_column=['label'],scale_type='z-score'):
        df_ret = df_data.copy(deep=True)
        if x_columns:
            scale_colums = x_columns
        else:
            scale_colums = []
            for key,type in zip(df_data.keys(),df_data.dtypes):
                if not type in ["bool","object","category",'datetime64','datetime'] and not key in y_column:
                    scale_colums.append(key)
                    if scale_type == 'z-score':
                        df_ret[[key]] = StandardScaler().fit_transform(df_ret[[key]])
                    elif scale_type == 'max-min':
                        df_ret[key] = (df_ret[key]-df_ret[key].min())/(df_ret[key].max()-df_ret[key].min())

                if type in ['bool']:
                    df_ret[key] = df_ret[key].astype(int)
                    
        #if not scale_colums:
        #    return df_ret,scale_colums
        #
        #if scale_type == 'z-score':
        #    scaler = StandardScaler()
        #    df_ret[scale_colums] = scaler.fit_transform(df_ret[scale_colums])
        #elif scale_type == 'max-min':
        #    for key in scale_colums:
        #        df_ret[key] = (df_ret[key]-df_ret[key].min())/(df_ret[key].max()-df_ret[key].min())
        return df_ret,scale_colums
        
    # 位置编码函数
    @staticmethod
    def get_transformer_position_encoding(seq_length, d_model):
        position_enc = np.array([
            [pos / np.power(10000, 2 * (i // 2) / d_model) for i in range(d_model)]
            for pos in range(seq_length)
        ])
        position_enc[:, 0::2] = np.sin(position_enc[:, 0::2])  # 偶数位置
        position_enc[:, 1::2] = np.cos(position_enc[:, 1::2])  # 奇数位置
        return position_enc
        
    # 等频率划分
    @staticmethod
    def assign_qbins(ds_data , quantiles ):
        ds_raw = ds_data.copy(deep = True)
        ds_noised = ds_data.copy(deep = True)
        scale = 10
        ds_scaled = ds_raw * scale
        epsilon = 1e-8
        ds_noised = ds_scaled + np.random.uniform(epsilon, epsilon*2, size=len(ds_scaled))
        ds_noised[ds_noised.idxmin()] = ds_scaled.min()

        ds_noised = ds_noised.drop_duplicates(keep='first')
        labels=range(quantiles)
        _, bin_edges  = pd.qcut(ds_noised,q=quantiles,duplicates='drop',retbins=True,precision=8)
        if len(labels) >= len(bin_edges):
            labels = labels[0:(len(labels)-1)]
        bins,bin_edges = pd.cut(ds_scaled, bins=bin_edges, labels=labels, include_lowest=True,retbins=True)
        bin_edges = bin_edges / scale
        intervals = [[bin_edges[i], bin_edges[i + 1]] for i in range(len(bin_edges) - 1)]
        return bins,intervals

class CFATransformer:
    
    def __init__(self, mode='copula', method='average', adjust_corr=False):
        """
        mode:
            'scale'     → 线性缩放（不改变分布）
            'quantile'  → 分位数映射（单变量/多变量）
            'copula'    → 多变量 copula 变换（默认）
        """
        self.mode = mode
        self.method = method
        self.adjust_corr = adjust_corr
        
        # 存储参数（用于 inverse）
        self.params = {}
    
    # ---------- 基础函数 ----------
    
    @staticmethod
    def _empirical_cdf(series, method='average'):
        ranks = series.rank(method=method)
        return (ranks - 0.5) / len(series)
    
    @staticmethod
    def _quantile_map(u, target):
        target_sorted = np.sort(target)
        return np.interp(
            u,
            np.linspace(0, 1, len(target_sorted)),
            target_sorted
        )
    
    # ---------- 核心接口 ----------
    
    def fit(self, source, target):
        """学习映射关系"""
        A = source
        B = target
        self.is_series = isinstance(A, pd.Series)
        
        if self.is_series:
            A = A.to_frame("col")
            B = B.to_frame("col")
        
        self.columns = A.columns
        
        # ---------- scale ----------
        if self.mode == 'scale':
            self.params['mean_A'] = A.mean()
            self.params['std_A'] = A.std()
            self.params['mean_B'] = B.mean()
            self.params['std_B'] = B.std()
        
        # ---------- quantile ----------
        elif self.mode == 'quantile':
            self.params['B_sorted'] = {
                col: np.sort(B[col].values) for col in self.columns
            }
        
        # ---------- copula ----------
        elif self.mode == 'copula':
            
            # B 的边际分布
            self.params['B_sorted'] = {
                col: np.sort(B[col].values) for col in self.columns
            }
            
            if self.adjust_corr:
                # B 的协方差
                B_u = B.apply(lambda x: self._empirical_cdf(x, self.method))
                B_z = pd.DataFrame(norm.ppf(B_u), columns=B.columns)
                
                cov_B = np.cov(B_z.T)
                cov_B += np.eye(len(cov_B)) * 1e-6
                
                self.params['L_B'] = np.linalg.cholesky(cov_B)
        
        return self
    
    # ---------- transform ----------
    
    def transform(self, data):
        A = data
        if self.is_series:
            A = A.to_frame("col")
        
        A_out = pd.DataFrame(index=A.index, columns=A.columns)
        
        # ---------- scale ----------
        if self.mode == 'scale':
            A_out = (A - self.params['mean_A']) / self.params['std_A']
            A_out = A_out * self.params['std_B'] + self.params['mean_B']
        
        # ---------- quantile ----------
        elif self.mode == 'quantile':
            for col in self.columns:
                u = self._empirical_cdf(A[col], self.method)
                A_out[col] = self._quantile_map(u, self.params['B_sorted'][col])
        
        # ---------- copula ----------
        elif self.mode == 'copula':
            
            # Step1: uniform
            A_u = A.apply(lambda x: self._empirical_cdf(x, self.method))
            
            # Step2: normal
            A_z = pd.DataFrame(norm.ppf(A_u), columns=A.columns)
            
            # Step3: adjust corr
            if self.adjust_corr:
                cov_A = np.cov(A_z.T)
                cov_A += np.eye(len(cov_A)) * 1e-6
                
                L_A = np.linalg.cholesky(cov_A)
                A_z = A_z @ np.linalg.inv(L_A).T @ self.params['L_B'].T
            
            # Step4: back to uniform
            A_u_new = pd.DataFrame(norm.cdf(A_z), columns=A.columns)
            
            # Step5: map to B
            for col in self.columns:
                A_out[col] = self._quantile_map(
                    A_u_new[col], self.params['B_sorted'][col]
                )
        
        if self.is_series:
            return A_out.iloc[:, 0]
        
        return A_out.astype(float)
    
    # ---------- inverse ----------
    
    def inverse(self, data_transformed, data_original):
        A_transformed = data_transformed
        A_original = data_original
        if self.is_series:
            A_transformed = A_transformed.to_frame("col")
            A_original = A_original.to_frame("col")

        A_rec = pd.DataFrame(index=A_transformed.index, columns=A_transformed.columns)
        
        # ---------- scale ----------
        if self.mode == 'scale':
            A_rec = (A_transformed - self.params['mean_B']) / self.params['std_B']
            A_rec = A_rec * self.params['std_A'] + self.params['mean_A']
        
        # ---------- quantile / copula ----------
        else:
            for col in self.columns:
                ranks = A_transformed[col].rank(method=self.method)
                u = (ranks - 0.5) / len(ranks)
                
                A_rec[col] = self._quantile_map(
                    u, A_original[col].values
                )
        
        if self.is_series:
            return A_rec.iloc[:, 0]
        
        return A_rec.astype(float)
    
def main():
    #正态转换
    ds_data = pd.Series([1,2,3,4,5,6])
    transformed_data = CFADataPreprocess.get_abnormal(ds_data)
    ds_data = ds_data.drop(transformed_data.index)
    #ds_data = ds_data[ds_data>0]
    ds_data.plot()
    plt.show()    

    #正态拟合
    p_object,e_std,v_object = CFADataPreprocess.polyfit(ds_y=ds_data,degree=3)
    print(v_object)
    
    #标准化
    df_sample = CFASample.get_random_classification(1000,n_feature=10,n_class=2)
    df_scale,scale_colums = CFADataPreprocess.get_scale(df_sample,y_column='y',scale_type="max-min")
    print(df_scale)

    #准备数据
    df_source = pd.DataFrame({'x': np.random.normal(0, 1, 100),'y': np.random.exponential(1, 100)})
    df_target = pd.DataFrame({'x': np.random.normal(10, 2, 200),'y': np.random.exponential(5, 200)})
    
    #单变量分布转换
    ds_source,ds_target = df_source['x'],df_target['x']
    df_result = pd.DataFrame()
    min_size = min(len(ds_source),len(ds_target))
    df_result['Source'] = ds_source.head(min_size)
    df_result['Target'] = ds_target.head(min_size)

    for mode in ['scale','quantile','copula']:
        transer = CFATransformer(mode=mode).fit(ds_source,ds_target)
        ds_tranformed = transer.transform(ds_source)
        ds_recovered = transer.inverse(ds_tranformed, ds_source)
        df_result[f"transfom_{mode}"] = ds_tranformed
        df_result[f"inverse_{mode}"] = ds_recovered
    
    print(df_result)

    #多变量分布转换
    min_size = min(len(df_source),len(df_target))
    df_result = pd.DataFrame()
    for key in df_source:
        df_result[f"source_{key}"] = df_source.head(min_size)[key]
        df_result[f"target_{key}"] = df_target.head(min_size)[key]

    for mode in ['scale','quantile','copula']:
        transer = CFATransformer(mode=mode).fit(df_source,df_target)
        df_tranformed = transer.transform(df_source)
        df_recovered = transer.inverse(df_tranformed, df_source)
        for key in df_tranformed:
            df_result[f"transfom_{mode}_{key}"] = df_tranformed[key]
            df_result[f"inverse_{mode}_{key}"] = df_recovered[key]
    print(df_result)

if __name__ == "__main__":         
    main()                         



#!/usr/bin/env bash
#!-*- coding:utf8-*-
import matplotlib.pyplot as plt
import seaborn as sns
import json,os
from pyecharts import options as opts
from pyecharts.charts import Sankey
import pandas as pd
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from pyvis.network import Network
from IPython.display import display, HTML  
from matplotlib.ticker import FuncFormatter
import webbrowser

def scientific(x, pos):
    # x:  tick value - ie. what you currently see in yticks
    # pos: a position - ie. the index of the tick (from 0 to 9 in this example)
    return '%.2E' % x
set_scientific = FuncFormatter(scientific)

class CFANetGraph:
    '''
    directed:True 有向图，False 无向图
    '''
    def __init__(self,directed = True):
        self.m_directed = directed
        self.m_nodes = []
        self.m_edges = []
        self.m_node_id = []
        
    def Add_Node(self,label,value,title,group=None,size=None,color='#00ff1e',node_id = None):
        tmp = {}
        if not node_id:
            node_id = len(self.m_nodes) + 1
        tmp['id'] = node_id
        tmp['label'] = label
        tmp['value'] = value
        tmp['title'] = title
        tmp['group'] = group
        tmp['size'] = size
        tmp['color'] = color
        if not node_id in self.m_node_id:
            self.m_nodes.append(tmp)
            self.m_node_id.append(node_id)
        return node_id
    
    def Add_Edge(self,from_id,to_id,weight,label):
        tmp = {}
        tmp['from'] = from_id
        tmp['to'] = to_id
        tmp['weight'] = weight
        tmp['label'] = label
        self.m_edges.append(tmp)
                
    def Show(self,file_name = 'pyvis.html'):
        net = Network('1024px','1024px',directed=self.m_directed)
        for node in self.m_nodes:
            net.add_node(node['id'],
                         label=node['label'],
                         value=node['value'],
                         color=node['color'],
                         title=node['title'],
                         group=node['group'])
        for edge in self.m_edges:
            net.add_edge(edge['from'], 
                         edge['to'], 
                         weight=edge['weight'],
                         label=edge['label'])
        net.toggle_physics(False)
        #net.show_buttons(filter_=['physics'])
        #net.show(file_name)
        net.write_html(file_name, notebook=False)
        #net.show(file_name)

    def Add(self,from_node,to_node,title=None,weight=None):
        node_id = None if 'id' not in from_node else from_node['id']
        label = None if 'label' not in from_node else from_node['label']
        value = None if 'value' not in from_node else from_node['value']
        title = None if 'title' not in from_node else from_node['title']
        group = None if 'group' not in from_node else from_node['group']
        size = None if 'size' not in from_node else from_node['size']
        color = '#00ff1e' if 'color' not in from_node else from_node['color']
        id_from = self.Add_Node(label=label,value=value,title=title,group=group,size=size,color=color,node_id=node_id)
       
        node_id = None if 'id' not in to_node else to_node['id']
        label = None if 'label' not in to_node else to_node['label']
        value = None if 'value' not in to_node else to_node['value']
        title = None if 'title' not in to_node else to_node['title']
        group = None if 'group' not in to_node else to_node['group']
        size = None if 'size' not in to_node else to_node['size']
        color = '#00ff1e' if 'color' not in to_node else to_node['color']
        id_to = self.Add_Node(label=label,value=value,title=title,group=group,size=size,color=color,node_id=node_id)
        
        self.Add_Edge(id_from,id_to,weight,title)
        return id_from,id_to

    @staticmethod
    def call_edge_info(from_node,to_node):
        weight = max(from_node['value'],to_node['value'])
        title = '%d:%d'%(from_node['value'],to_node['value'])
        return weight,title
    '''
    可视化时序数据
    df_in:时序数据，必须的字段有label(节点的标签),value(节点的值),title(节点的tips)
    call_edge_info: callback根据from和to节点，计算edge的title和weight
    file_name：输出的html文件
    '''
    @staticmethod
    def Show_Series(df_in,call_edge_info = None,file_name = 'pyvis.html'):
        net = CFANetGraph()
        for item in json.loads(df_in.to_json(orient='records')):
            label = None if 'label' not in item else item['label']
            value = None if 'value' not in item else item['value']
            title = None if 'title' not in item else item['title']
            group = None if 'group' not in item else item['group']
            size = None if 'size' not in item else item['size']
            color = '#00ff1e' if 'color' not in item else item['color']
            net.Add_Node(label=label,value=value,title=title,group=group,size=size,color=color)
   
        for i in range(df_in.shape[0]-1):
            curr_node = json.loads(df_in.iloc[i].to_json())
            next_node = json.loads(df_in.iloc[i+1].to_json())
            if call_edge_info:
                weight,title = call_edge_info(curr_node,next_node)
            else:
                weight,title = CFANetGraph.call_edge_info(curr_node,next_node)
            net.Add_Edge(i+1,i+2,weight,title)
            
        file_path = os.path.abspath(file_name)
        net.Show(file_name)
        webbrowser.open(f"file://{file_path}")  # 自动用默认浏览器打开

class CFAVisualize:
    
    def __init__(self):
        pass

    '''
    绘制热图
    '''
    @staticmethod
    def show_heatmap(df,title="heatmap demo"):
        #plt.ticklabel_format(style='plain', axis='both')
        sns.set(font_scale=1.5)
        cmap = sns.cubehelix_palette(start=1, rot=3, gamma=0.8, as_cmap=True)
        sns.heatmap(df, cmap=cmap, linewidths=0.05, annot=False, fmt="g", annot_kws={"fontsize": 32})
        plt.xticks(rotation=60)
        plt.yticks(rotation=15)
    '''
    绘制桑基图
    1.节点，边，流量
    2.nodes中的"name"不要重命, 否则会不识别, 导致没有图
    3.links中source和target对应的值一定要在nodes的name对应的值中, 否则图中会不显示(但是不会报错)
    4.links中source和target对应的值不能相同, 否则图不显示
    '''
    @staticmethod
    def get_sankey(df_data_list,title="sankey demo"):
        if type(df_data_list) == list:
            return CSHVisualize.get_sankey_base(df_data_list,title)
        else:
            df_tmp =  df_data_list
            columns = df_tmp.columns.tolist()
            df_list = []    
            value_key = columns[-1]
            for i in range(1,len(columns) - 1):
                groupList = []
                name1 = columns[i-1]
                name2 = columns[i]
                for j in range(0,i+1):
                    name = columns[j]
                    groupList.append(name)
                df_t = df_tmp.groupby(groupList).sum()[value_key].reset_index()
                df_values = pd.DataFrame()
                df_values[name1] = df_t[name1]
                df_values[name2] = df_t[name2]
                df_values[value_key] = df_t[value_key]
                df_list.append(df_values)
                
            return CFAVisualize.get_sankey_base(df_list,title)
        
    @staticmethod
    def get_sankey_base(df_data_list,title="sankey demo"):
        def get_sankey_data(df_in):
            def get_name(feild,value):
                #return str(value)
                return str(feild)+":"+str(value)
            #create nodes
            columns = df_in.columns.tolist()
            nodes = []
            for item in df_in[columns[0]].unique():
                dic = {}
                dic['name'] = get_name(columns[0],item)
                nodes.append(dic)
            for item in df_in[columns[1]].unique():
                dic = {}
                dic['name'] = get_name(columns[1],item)
                nodes.append(dic)
            #create links
            links = []
            for item in df_in.values:
                dic = {}
                dic['source'] = get_name(columns[0],item[0])
                dic['target'] = get_name(columns[1],item[1])
                dic['value'] = float(item[2])
                links.append(dic)
            df_tmp = pd.DataFrame(nodes)
            #df_tmp = df_tmp.sort_values(by='value',ascending=False)
            nodes = json.loads(df_tmp.to_json(orient='records'))
            return nodes,links
        nodes,links = [],[]
        if type(df_data_list) != list:
            nodes,links = get_sankey_data(df_data_list)
        else:         
            for df in df_data_list:
                nodes1,links1 = get_sankey_data(df)
                for n in nodes1:
                    if not n in nodes:
                        nodes.append(n)
                for l in links1:
                    links.append(l)
                
        graph = Sankey(init_opts=opts.InitOpts(width="1024", height="768px",theme='westeros'))
        graph.add(title,nodes=nodes,links=links,
              linestyle_opt=opts.LineStyleOpts(opacity=0.2, curve=0.5, color="source"),
              label_opts=opts.LabelOpts(position="right"))
        return graph
    '''
    显示等高线
    df_data：待显示数据，格式为（x，y，value）
    isBlack: 黑白显示
    withTips：是否显示数字
    '''
    @staticmethod
    def show_contour(df_data, isBlack=True, withTips = True):
        columns = df_data.columns.tolist()
        df_tmp = df_data.copy(deep = True)
        category_map = {}
        for i in range(2):
            key = columns[i]
            key_type = str(df_tmp[key].dtypes)
            if key_type == "object":
                df_tmp[key] = df_tmp[key].astype("category")
                category_map[key] = dict(zip(df_tmp[key].cat.codes,df_tmp[key]))            
            key_type = str(df_tmp[key].dtypes)
            if key_type == "category":
                df_tmp[key] = df_tmp[key].cat.codes
                
        hdfpivot = df_tmp.pivot(index=columns[0], columns=columns[1], values=columns[2])
        X = hdfpivot.columns.values
        Y = hdfpivot.index.values
        Z = hdfpivot.values
        x, y = np.meshgrid(X, Y)
        #plt.ticklabel_format(style='plain', axis='both')
        cmap = sns.cubehelix_palette(start=1, rot=3, gamma=0.8, as_cmap=True)
        if isBlack:
            contours = plt.contour(x, y, Z, 3, colors='black')  # the NAN will be plotted as white spaces
        else:
            contours = plt.contourf(x, y, Z ,cmap=cmap, linewidths=0.05 ) #the NAN will be plotted as white spaces
        if withTips:
            plt.clabel(contours, inline=True, fontsize=8)
        plt.colorbar()
        return category_map
    '''
    可视化时序数据
    df_in:时序数据，必须的字段有label(节点的标签),value(节点的值),title(节点的tips)
    call_edge_info: callback根据from和to节点，计算edge的title和weight
    file_name：输出的html文件
    '''
    @staticmethod
    def show_sequence(df_data, call_edge_info = None,file_name = 'pyvis.html'):
        CFANetGraph().Show_Series(df_data,call_edge_info,file_name)

def test_heatmap():
    data_list = []
    data_list.append({"类别1": "A1", "类别2": "B1", "Count": 100})
    data_list.append({"类别1": "A2", "类别2": "B2", "Count": 50})
    data_list.append({"类别1": "A3", "类别2": "B1", "Count": 50})
    data_list.append({"类别1": "A3", "类别2": "B2", "Count": 10})
    df_data = pd.DataFrame(data_list)
    pt = df_data.pivot_table(values='Count', index=["类别1"], columns=['类别2'], aggfunc=np.sum, fill_value=0)
    CFAVisualize.show_heatmap(pt)

def test_sankey():
    data_list = []
    data_list.append({"类别1": "A1", "类别2": "B1", "Count": 100})
    data_list.append({"类别1": "A2", "类别2": "B2", "Count": 50})
    data_list.append({"类别1": "A3", "类别2": "B1", "Count": 50})
    data_list.append({"类别1": "A3", "类别2": "B2", "Count": 10})
    df_data = pd.DataFrame(data_list)
    Sankey = CFAVisualize.get_sankey(df_data)
    html_file = './my_chart.html'
    file_path = os.path.abspath("my_chart.html")
    Sankey.render(file_path)  # 生成 HTML 文件
    webbrowser.open(f"file://{file_path}")

def test_contour():
    data_list = []
    for i in range(10):
        for j in range(10):
            tmp = {"x": i, "y": j, "Count": np.sin(i * j)}
            data_list.append(tmp)
    df_data = pd.DataFrame(data_list)
    CFAVisualize.show_contour(df_data)
    
def test_sequence():
    data = []
    for i in range(5):
        tmp = {}
        tmp['label'] = "name-%d"%i
        tmp['value'] = i
        tmp['title'] = "title-%d"%i

        data.append(tmp)
    df_data = pd.DataFrame(data)
    CFAVisualize.show_sequence(df_data)
    
def main():
    test_heatmap()
    plt.show()

    test_contour()
    plt.show()

    test_sankey()
    test_sequence()

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# coding: utf-8
#!/usr/bin/env bash
import pandas as pd
import numpy as np
from tqdm import tqdm
import subprocess

class CFACommon:
    def __init__(self):
        pass
    '''
    带进度条，读取csv文件
    '''
    @staticmethod
    def load_csv(file_name,chunksize=1000):
        out = subprocess.getoutput("wc -l %s" % file_name)
        total = int(out.split()[0]) / chunksize
        return pd.concat([chunk for chunk in tqdm(pd.read_csv(file_name, chunksize=chunksize),total=total, desc='Loading data %s'%file_name)])
    
def main():
    pass

if __name__ == "__main__":
    main()






'''
特征选择
1. CFAFeatureSelect
   --生成信息图(分类数据)
   --GLM-ANOVA 方差检验（回归数据）
   --格兰特因果检验（时序数据）
'''
import h2o,json,os,sys,warnings
from datetime import datetime
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import pandas as pd
from tqdm import tqdm
import os,sys
sys.path.append(os.path.dirname(os.path.dirname(__file__)))
from FreeAeonML.FACommon import CFACommon
from FreeAeonML.FASample import CFASample
from h2o.automl import H2OAutoML
from h2o.estimators import *
from statsmodels.tsa.stattools import grangercausalitytests
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from mpl_toolkits.mplot3d import Axes3D
from sklearn.cluster import KMeans

np.set_printoptions(suppress=True)
pd.set_option('display.float_format',lambda x : '%.8f' % x)

class CFAFeatureSelect():
 
    def __del__(self):
       pass
       #h2o.shutdown()

    def __init__(self,ip='localhost',port=54321):
        #h2o.init(nthreads = -1, verbose=False)
        self.m_df_h2o = None
        self.m_col_x = []
        self.m_col_y = None

    def load(self,df_sample,x_columns = [],y_column='label',is_regression = False):
        
        if x_columns == []:
            total_columns = df_sample.keys().tolist()
        else:
            total_columns = x_columns
         
        if y_column in total_columns:
            total_columns.remove(y_column)
                
        self.m_col_x = total_columns
        self.m_col_y = y_column
        
        self.m_df_h2o = h2o.H2OFrame(df_sample)
        if not is_regression:
            self.m_df_h2o[self.m_col_y] = self.m_df_h2o[self.m_col_y].asfactor()
    '''
    横坐标：信息总量（total information）
        --变量对预测的影响，,即该变量与其他变量的相关性。
        --横轴上的值越大，表示变量对响应的影响越显著。
    纵坐标：净信息（net information）
        --变量的独特性，总信息量.
        --净信息越高，预测能力越强，表示该变量对响应的影响越独特。
    可接受特征
        --位于虚线以上和右侧的特征是最具预测能力和独特性的特征，它们被认为是可接受的特征，
        --这些特征是被认为是核心驱动因素的变量，它们在总信息（预测能力）和净信息（独特性）方面都表现出色。        
        --可以用于建立模型和做出决策。
    返回值方法：
    ig.get_admissible_score_frame()
    ig.get_admissible_features()
    '''
    def get_inform_graph(self,algorithm="AUTO",protected_columns=[]): #["All",'AUTO','deeplearning','drf','gbm','glm','xgboost']
        if algorithm == "All":
            ret = {}
            for algor in ['AUTO','deeplearning','drf','gbm','glm']: #,'xgboost']: xgboost 有 bug
                if protected_columns:
                    ig = H2OInfogram(algorithm=algor,protected_columns=protected_columns)
                else:
                    ig = H2OInfogram(algorithm=algor)
                ig.train(x=self.m_col_x, y=self.m_col_y,training_frame=self.m_df_h2o)
                ret[algor] = ig
            return ret
        else:
            if protected_columns:
                ig = H2OInfogram(algorithm=algorithm,protected_columns=protected_columns)
            else:
                ig = H2OInfogram(algorithm=algorithm)
            ig.train(x=self.m_col_x, y=self.m_col_y,training_frame=self.m_df_h2o)
            return ig
    '''
    统计自变量和因变量的相关性
        --p值小于0.05,认为有相关性
        --通过特征组集合，查看特征之间是否有相关性
    '''           
    def get_anovaglm(self,family='gaussian',lambda_ = 0,highest_interaction_term=2):
        anova_model = H2OANOVAGLMEstimator(family=family,
                                   lambda_=lambda_,
                                   missing_values_handling="skip",
                                   highest_interaction_term=highest_interaction_term)
        anova_model.train(x=self.m_col_x, y=self.m_col_y, training_frame=self.m_df_h2o)
        return anova_model
        #anova_model.summary()

    '''
    格兰特因果检验
    ds_result：结果序列（必须为平稳序列）
    ds_source：原因序列（必须为平稳序列）
    maxlag：时间间隔（整数或列表，为整数时，遍历所有的lag）
    返回值：
    1. 最小的p值
    2. 最佳lag
    3. approve_list通过测试的lag
    4. 详细测试结果
    '''
    @staticmethod
    def granger_test(ds_result,ds_source,maxlag,p_value = 0.05):

        df_test = pd.DataFrame()
        df_test['result'] = ds_result
        df_test['source'] = ds_source

        gc_result = grangercausalitytests(df_test[['result', 'source']], maxlag=maxlag,verbose=False )
        min_lag_p = 100
        best_lag = -1
        detail = {}
        approve_list = []
        for lag in gc_result:

            result = gc_result[lag][0]
            detail[lag] = result.copy()
    
            max_p = 0 
            for test in result:
                test_value = result[test][0]
                test_p = result[test][1]
                if test_p > max_p:
                    max_p = test_p

            if max_p > p_value:
                continue
            #if max_p == 0:
            #    continue   
            approve_list.append({'lag':lag,"max p-value":max_p})
            
            if min_lag_p > max_p:
                min_lag_p = max_p
                best_lag = lag 

        return min_lag_p, best_lag, approve_list, detail   
    
    '''
    使用PCA，得到降维后和重构后的矩阵（根据特征值和特征向量）
    dataMat：原始矩阵
    n：特征向量的维度（降维后）
    '''
    @staticmethod
    def get_matrix_by_pca(dataMat, n):
        # 零均值化
        def zeroMean(dataMat):
            meanVal = np.mean(dataMat, axis=0)  # 按列求均值，即求各个特征的均值
            newData = dataMat - meanVal
            return newData, meanVal

        newData, meanVal = zeroMean(dataMat)
        covMat = np.cov(newData, rowvar=0)  # 求协方差矩阵,return ndarray；若rowvar非0，一列代表一个样本，为0，一行代表一个样本
        eigVals, eigVects = np.linalg.eig(np.asmatrix(covMat))  # 求特征值和特征向量,特征向量是按列放的，即一列代表一个特征向量
        # argsort将x中的元素从小到大排列，提取其对应的index(索引)
        eigValIndice = np.argsort(eigVals)  # 对特征值从小到大排序
        # print(eigValIndice)
        n_eigValIndice = eigValIndice[-1:-(n + 1):-1]  # 最大的n个特征值的下标
        n_eigVect = eigVects[:, n_eigValIndice]  # 最大的n个特征值对应的特征向量
        lowDDataMat = newData * n_eigVect  # 低维特征空间的数据
        reconMat = (lowDDataMat * n_eigVect.T) + meanVal  # 重构数据
        return lowDDataMat, reconMat
    
    '''
    使用PCA和TSNE，得到降维后和重构后的矩阵（根据特征值和特征向量）
    返回PCA降维后结果df_pca,和STNE降维后的结果df_tsne
    '''
    @staticmethod
    def get_data_pca(df_samples,n_components=2,label_column='y',feature_list=[],with_plot=True,perplexity=None,n_clusters = 2):
        if feature_list:
            X = df_samples[feature_list] 
        else:
            features = df_samples.keys().tolist()
            if label_column in features:
                features.remove(label_column)
            X = df_samples[features]

        if label_column in df_samples.keys():
            y = df_samples[label_column]
        else:
            y = None

        pca = PCA(n_components=n_components)
        X_pca = pca.fit_transform(X)
        
        if perplexity == None:
            perplexity = min(30, X.shape[0] // 30)

        if y is None:
            kmeans = KMeans(n_clusters=n_clusters, random_state=42)
            y = kmeans.fit_predict(X_pca)  # 聚类结果作为颜色

        tsne = TSNE(n_components=n_components, random_state=42, perplexity=perplexity)
        X_tsne = tsne.fit_transform(X)
        if with_plot:
            fig, axs = plt.subplots(1, 2, figsize=(12, 5))
            if n_components == 2:
                axs[0].scatter(X_pca[:, 0], X_pca[:, 1], c=y, cmap='turbo', s=10)
                axs[0].set_title('PCA (2D)')
                axs[0].axis('on')
                axs[0].grid(True, linestyle='--', alpha=0.5) 

                axs[1].scatter(X_tsne[:, 0], X_tsne[:, 1], c=y, cmap='turbo', s=10)
                axs[1].set_title('t-SNE (2D)')
                axs[1].axis('on')
                axs[0].grid(True, linestyle='--', alpha=0.5)

            elif n_components == 3:
                ax1 = fig.add_subplot(121, projection='3d')
                ax1.scatter(X_pca[:, 0], X_pca[:, 1], X_pca[:, 2], c=y, cmap='turbo', s=10)
                ax1.set_title('PCA (3D)')
                ax1.axis('on')
                ax1.grid(True, linestyle='--', alpha=0.5) 
                
                ax2 = fig.add_subplot(122, projection='3d')
                ax2.scatter(X_tsne[:, 0], X_tsne[:, 1], X_tsne[:, 2], c=y, cmap='turbo', s=10)
                ax2.set_title('t-SNE (3D)')
                ax2.axis('on')
                ax2.grid(True, linestyle='--', alpha=0.5) 

            plt.tight_layout()
            plt.show()
        
        df_pca,df_tsne = pd.DataFrame(X_pca),pd.DataFrame(X_tsne)

        if label_column in df_samples.keys():
            df_pca[label_column] = df_samples[label_column]
            df_tsne[label_column] = df_samples[label_column]

        return df_pca,df_tsne

def main():
    
    #如果是WSL,注释掉h2o.init(),使用h2o.connect()
    h2o.init(nthreads = -1, verbose=False)
    #h2o.connect(ip=ip,port=port)

    #分类数据,查看信息图
    df_sample = CFASample.get_random_classification(1000,n_feature=10,n_class=2)
    Fea = CFAFeatureSelect()
    Fea.load(df_sample,x_columns = ['x1','x2'],y_column='y',is_regression = False)
    ig = Fea.get_inform_graph("AUTO")
    ig.plot()
    ig.show()

    #回归数据，查看方差检验
    df_sample = CFASample.get_random_regression()
    Fea.load(df_sample,x_columns = ['x1','x2'],y_column='y',is_regression = True)
    ag = Fea.get_anovaglm()
    print(ag.summary())

    #格兰特因果检验
    a = [1,-1,2,-2,3,-3.1]
    b = [2,-2,3,-3,4,-4.1]
    result = CFAFeatureSelect.granger_test(b,a,[1])
    print(result)
        
    a.extend(a)
    b.extend(b)
    result = CFAFeatureSelect.granger_test(b,a,2)
    print(result)


    # 生成示例数据 500个样本，5个特征
    data = np.random.rand(500, 5)
    labels = np.random.randint(0, 2, size=500)
    df_data = pd.DataFrame(data)
    df_data['y'] = labels
    df_pca,df_sne = CFAFeatureSelect.get_data_pca(df_data,n_components=2,label_column='y')
    print(df_pca)
    print(df_sne)
    
    # 降维到2维
    lowDData, reconData = CFAFeatureSelect.get_matrix_by_pca(data, 2)

    print("降维后的数据形状:", lowDData.shape)  # (10, 2)
    print("重构数据形状:", reconData.shape)      # (10, 5)

    print("降维后的数据:\n", lowDData)
    print("重构后的数据:\n", reconData)

if __name__ == "__main__":
    main()



'''
自动化训练分类模型（支持多分类），包括
1.模型训练
2.模型评估
3.ROC曲线绘制
'''
import platform
import numpy as np
import pandas as pd
import h2o,json,os,sys,time,datetime,warnings
from datetime import datetime
import matplotlib.pyplot as plt
import seaborn as sns
from tqdm import tqdm
from h2o.automl import H2OAutoML
from h2o.estimators import *
from sklearn.neighbors import KNeighborsClassifier
from sklearn.datasets import make_regression,make_classification
from sklearn.metrics import recall_score,confusion_matrix,matthews_corrcoef,accuracy_score,precision_score,roc_auc_score,log_loss,f1_score,roc_curve,fbeta_score
from sklearn.metrics import confusion_matrix
import pickle
import matplotlib.pyplot as plt
import seaborn as sns
sys.path.append(os.path.dirname(os.path.dirname(__file__)))
from FreeAeonML.FACommon import CFACommon
from FreeAeonML.FASample import CFASample
np.set_printoptions(suppress=True)
pd.set_option('display.float_format',lambda x : '%.8f' % x)

'''
使用h2o进行automl
'''
class CFAModelClassify():
    
    def __init__(self,ip='localhost',port=54321,models = None):
        #h2o.init(nthreads = -1, verbose=False)
        #h2o.connect(ip=ip,port=port,verbose=False)
        if models == None:
            self.m_models = {
                "dt":H2ODecisionTreeEstimator(),
                "svm":H2OSupportVectorMachineEstimator(), # kernel='gaussian'
                "rf":H2ORandomForestEstimator(),
                "ann":H2ODeepLearningEstimator(hidden=[100, 100]),
                "bayes":H2ONaiveBayesEstimator(),
                "glm":H2OGeneralizedLinearEstimator(),
                "gbm":H2OGradientBoostingEstimator()
                #"xgboost":H2OXGBoostEstimator()
            }
            # 判断平台与可用性
            system = platform.system()
            machine = platform.machine()

            try:
                if H2OXGBoostEstimator.available():
                    if system != "Windows" and not machine.startswith("arm"):
                        self.m_models["xgboost"] = H2OXGBoostEstimator()
                    else:
                        print("⚠️ 当前平台不支持 H2O XGBoost（Windows 或 ARM 架构）")
                else:
                    print("⚠️ H2O XGBoostEstimator 不可用（编译未启用或未安装）")
            except Exception as e:
                print(f"⚠️ 检查 H2O XGBoost 可用性失败：{e}")

        else:
            self.m_models = models

    def train(self,df_sample,x_columns = [] ,y_column='label',train_ratio = 0.85):
        
        if x_columns == []:
            total_columns = df_sample.keys().tolist()
        else:
            total_columns = x_columns
            if not y_column in total_columns:
                total_columns.append(y_column)
            
        df_temp = df_sample[total_columns]
        
        df_h2o = h2o.H2OFrame(df_temp)
        if train_ratio > 0:
            df_train, df_valid = df_h2o.split_frame(ratios=[train_ratio], seed=1234)
        else:
            df_train = df_h2o
            
        x = df_train.columns
        y = y_column
        x.remove(y)
        df_train[y] = df_train[y].asfactor()
        if train_ratio > 0:
            df_valid[y] = df_valid[y].asfactor()
        
        for key in self.m_models:
            model = self.m_models[key]
            #print("begin train ",key)
            if train_ratio > 0:
                model.train(x=x, y=y,training_frame=df_train,validation_frame=df_valid)
            else:
                model.train(x=x, y=y,training_frame=df_train)
            #print("end train ",key)
    
    def load(self,model_path):
        for key in self.m_models:
            folder = "%s/%s"%(os.path.abspath(model_path),key)
            model = self.m_models[key]
            all_models = os.listdir(folder)
            all_models.sort()
            for file_name in all_models:
                model_file = os.path.join(folder, file_name)
                print("loading",model_file)
                self.m_models[key] = h2o.load_model(model_file)
   
    def save(self,model_path):
        for key in self.m_models:
            folder = "%s/%s"%(os.path.abspath(model_path),key)
            model = self.m_models[key]
            if not os.path.exists(folder):
                os.makedirs(folder)
                
            for file_name in os.listdir(folder):
                model_file = os.path.join(folder, file_name)
                os.remove(model_file)
            
            model_saved = h2o.save_model(model=model, path=folder, force=True)
            print("save model to ",model_saved)

    def predict(self,df_sample,x_columns = [],y_column='label'):
        
        if x_columns == []:
            total_columns = df_sample.keys().tolist()
        else:
            total_columns = x_columns
         
        if y_column in total_columns:
            total_columns.remove(y_column)
                
        if y_column in df_sample.keys().tolist():
            y_true = df_sample[y_column].tolist()
        else:
            y_true = None
            
        df_temp = df_sample[total_columns]
        df_h2o = h2o.H2OFrame(df_temp)

        result = []
        for key in self.m_models:
            model = self.m_models[key]
            training_columns = model._model_json['output']['names'][:-1]
            for col in training_columns:
                if col not in df_temp.columns:
                    print("NO training column %s found not"%col)
                    df_temp = df_temp.reindex(columns=training_columns, fill_value=0)  
                    df_h2o = h2o.H2OFrame(df_temp)
                    break
            df_t = model.predict(df_h2o).as_data_frame()
            df_t['model'] = key
            df_t['true'] = y_true
            result.extend(json.loads(df_t.to_json(orient='records')))
            
        return pd.DataFrame(result)
    
    #average is [None, 'micro', 'macro', 'weighted']
    def evaluate(self,df_sample,x_columns = [],y_column='label',average='weighted'):
        
        df_pred = self.predict( df_sample,x_columns=x_columns , y_column=y_column )
        prob_list = []
        for i in range( df_sample[y_column].nunique() ):
            prob_list.append("p%d"%i)  
        result = []
        for model,df in df_pred.groupby('model'):
            y_pred = df['predict'].tolist()
            y_true = df['true'].tolist()
            for key in prob_list:
                if not key in df.keys().to_list():
                    df[key] = 0
            y_prab = df[prob_list].to_numpy()
            tmp = {}
            tmp['model'] = model
            tmp['confusion_matrix'] = confusion_matrix(y_true,y_pred)
            tmp['recall'] = recall_score(y_true,y_pred,average=average)
            tmp['mcc'] = matthews_corrcoef(y_true,y_pred)
            tmp['accuracy'] = accuracy_score(y_true,y_pred)
            tmp['precision'] = precision_score(y_true,y_pred,average=average)
            if df['true'].nunique() > 2:
                tmp['auc_ovo'] = roc_auc_score(y_true,y_prab,multi_class='ovo',average=average)
                tmp['auc_ovr'] = roc_auc_score(y_true,y_prab,multi_class='ovr',average=average)
            else:
                tmp['auc'] = roc_auc_score(y_true,y_pred,average=average)
            tmp['f1_score'] = f1_score(y_true,y_pred,average=average)
            tmp['fbeta_score'] = fbeta_score(y_true,y_pred,beta=0.5,average=average)
            
            result.append(tmp)
            
        return pd.DataFrame( result )
    
    def importance(self):
        result = []
        for key in self.m_models:
            model = self.m_models[key]
            df_temp = model.varimp(use_pandas=True)
            if type(df_temp) == pd.DataFrame:
                df_temp['model'] = key
                result.extend(json.loads(df_temp.to_json(orient='records')))
        df_importance = pd.DataFrame(result).reset_index(drop=True)
        return df_importance

def main():

    #如果是WSL,注释掉h2o.init(),使用h2o.connect()
    h2o.init(nthreads = -1, verbose=False)
    #h2o.connect(ip=ip,port=port)

    df_sample = CFASample.get_random_classification(1000,n_feature=1,n_class=2)
    df_train,df_test = CFASample.split_dataset(df_sample)
    print(df_train)

    model_1 = CFAModelClassify()
    model_1.train(df_train,y_column='y')
    model_1.save("./test")
    model_2 = CFAModelClassify()
    model_2.load("./test")
    df_pred = model_2.predict(df_test,y_column='y')
    print(df_pred)
    df_evaluate = model_2.evaluate(df_test,y_column='y')
    print(df_evaluate)
    df_importance = model_2.importance()
    print(df_importance)
    nan_rows = df_pred[df_pred.isna().any(axis=1)]
    print(nan_rows)
    df_pred = df_pred.dropna()
    print(df_pred)

if __name__ == "__main__":
    main()



import pandas as pd
import numpy as np
from sklearn.datasets import make_regression
from sklearn.datasets import make_classification
from sklearn.model_selection import train_test_split
#from sklearn.datasets.samples_generator import make_blobs
from sklearn.datasets import make_blobs
from imblearn.over_sampling import SMOTE,SMOTENC

np.set_printoptions(suppress=True)
pd.set_option('display.float_format',lambda x : '%.8f' % x)

class CFASample():
    
    @staticmethod 
    def get_random_regression(n_sample = 100):
        df = pd.DataFrame()
        features, target = make_regression(n_samples=n_sample,
                                           n_features=2,
                                           n_targets=1,
                                           random_state=100,
                                           noise = 1)
        df['x1'] = features[:,0]
        df['x2'] = features[:,1]
        df['y'] = features[:,0] + features[:,1]
        return df
    
    @staticmethod 
    def get_random_classification(n_sample = 100 , n_feature =2, n_class = 2):
        df = pd.DataFrame()
        X, y = make_classification(n_samples=n_sample,
                                   n_features=n_feature,
                                   n_redundant=0,
                                   n_classes=n_class, # binary target/label
                                   n_clusters_per_class = 1,
                                   n_informative = n_feature,
                                   flip_y=0.1,  #high value to add more noise
                                   random_state=100)
        for i in range(n_feature):
            df['x%d'%i] = X[:,i]
        
        df['y'] = y
        return df

    @staticmethod
    def get_random_cluster(n_sample = 100):
        centers = [(-5, -5), (5, 5)]
        cluster_std = [0.8, 1]
        X, y = make_blobs(n_samples=n_sample , cluster_std=cluster_std, centers=centers, n_features=2, random_state=1)
        df = pd.DataFrame()
        df['x1'] = X[:,0]
        df['x2'] = X[:,1]
        df['y'] = y
        return df

    @staticmethod
    def split_dataset(df,test_ratio = 0.2):
        df_train,df_test = train_test_split(df, test_size = test_ratio)
        return df_train.reset_index(drop=True),df_test.reset_index(drop=True)

    @staticmethod
    def resample_smote(df_sample,x_columns=[],y_column='label',random_state=42):

        if x_columns == []:
            total_columns = [col for col in df_sample.columns if col != y_column]
        else:
            total_columns = x_columns

        total_columns = [col for col in df_sample.columns if col != y_column]
        categorical_features = [
            i for i, col in enumerate(total_columns)
            if df_sample[col].dtype == 'object'
        ]
        X = df_sample[total_columns]
        y = df_sample[y_column]
        if categorical_features:
            smote_nc = SMOTENC(categorical_features=categorical_features, random_state=random_state)
            X_resampled, y_resampled = smote_nc.fit_resample(X, y)
        else:
            smote = SMOTE(random_state=random_state)
            X_resampled, y_resampled = smote.fit_resample(df_sample[total_columns], df_sample[y_column])

        df_ret = pd.DataFrame(X_resampled, columns=total_columns)
        df_ret[y_column] = y_resampled

        return df_ret
    
    @staticmethod
    def resample_balance(df_data,y_column = 'labels'):
        df_minority = df_data[df_data[y_column] == df_data[y_column].value_counts().idxmin()]
        df_majority = df_data[df_data[y_column] == df_data[y_column].value_counts().idxmax()]
        df_minority_upsampled = df_minority.sample(len(df_majority), replace=True)
        df_balanced = pd.concat([df_minority_upsampled, df_majority])
        df_balanced = df_balanced.sample(frac=1).reset_index(drop=True)
        return df_balanced
    
    '''
    找到序列中，发生变化的index
    '''
    @staticmethod
    def find_changed_index(ds):
        df = pd.DataFrame()
        ds_1 = ds.fillna(1e-11)  # 用极小值填充 NaN
        ds_1 = pd.concat([ds_1, pd.Series([ds.iloc[-1]], index=[ds.index[-1] + 1])])
        ds_2 = ds_1.shift(1)
        df['x'] = ds_1
        df['y'] = ds_2
        df=df[0:-1]
        return df[df['x']!=df['y']].index.values
    
    '''
    将DataFrame划分成多个batch
    '''
    @staticmethod
    def split_dataframe(df_sample,batch_size = 10):
        return [df_sample[i:i + batch_size] for i in range(0, len(df_sample), batch_size)]
        
def main():
    df_classification = CFASample.get_random_classification(1000,n_feature=10,n_class=4)
    df_regression = CFASample.get_random_regression()
    df_cluster = CFASample.get_random_cluster()
    print(df_classification)
    print(df_regression)
    print(df_cluster)

    df_train,df_test = CFASample.split_dataset(df_regression)
    print(df_train.shape)
    print(df_test.shape)

    df_sample = CFASample.resample_smote(df_classification,y_column='y')
    df_sample = CFASample.resample_balance(df_classification,y_column='y')

    print(df_sample)

if __name__ == "__main__":
    main()




'''
数据探索
1. CFADataDistribution (数据分布)
    --显示直方图、KED图和Q-Q图
    --检查是否为高斯分布
    --比较两个分布是否相私
2. CFADataTest（假设检验）
    --检验是否为平稳序列
    --显示自相关和偏自相关图
3. CFAFitter (数据拟合)
    --linear:线性拟合
    --polynomial:多项式拟合
    --exponential:指数拟合
4. CFACommonStats (计算常见统计量)
    --mean,std,cv
    --skew,kurt_fisher,kurt_pearson,autocorr_lag1
    --min,max,range,iqr

'''
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.stats import probplot,shapiro,kstest,normaltest,chisquare
import seaborn as sns
import json
from scipy.stats import f_oneway,norm, expon
from statsmodels.tsa.stattools import adfuller 
from statsmodels.graphics.tsaplots import plot_acf, plot_pacf
from scipy.optimize import curve_fit
from statsmodels.tsa.stattools import grangercausalitytests

class CFADataDistribution:
    def __init__(self):
        pass
    '''
    显示分布图
    '''
    @staticmethod
    def show_dist(ds,bins = 10 ):
        fig, ax =plt.subplots(1,3,constrained_layout=True, figsize=(8, 3))
        ds.hist(ax=ax[0],bins=bins)
        ax[0].set_title("Histogram")
        ds.plot(kind="kde",ax=ax[1])
        ax[1].set_title("KDE")
        probplot(x=ds,dist='norm',plot=ax[2])
        ax[2].set_title("Q-Q")
        plt.show()
   
    '''
    检查是否为高斯分布
    '''
    @staticmethod
    def normal_test(ds , p_value = 0.05):
        ret = {}
        ret['shapiro-wilk'] = {}
        temp = shapiro(ds)
        ret['shapiro-wilk']['is_normal'] = str(temp[1] >= p_value)
        ret['shapiro-wilk']['detail'] = temp
        ret['shapiro-wilk']['describe'] = "Shapiro-Wilk检验，适用于小样本场合（3≤n≤50），受异常值影响较大。"
        ret['kolmogorov-smirnov'] = {}
        temp = kstest(ds,"norm")
        ret['kolmogorov-smirnov']['is_normal'] = str(temp[1] >= p_value)
        ret['kolmogorov-smirnov']['detail'] = temp
        ret['kolmogorov-smirnov']['describe'] = "Kolmogorov-Smirnov检验是一项拟合优度的统计检验。 此测试比较两个分布（在这种情况下，两个分布之一是高斯分布）。 此检验的零假设是，两分布相同（或），两个分布之间没有差异。"
        ret['skewness-kurtosis'] = {}
        temp = normaltest(ds,)
        ret['skewness-kurtosis']['is_normal'] = str(temp[1] >= p_value)
        ret['skewness-kurtosis']['detail'] = temp
        ret['skewness-kurtosis']['describe'] = "DAgostino-Pearson方法使用偏度和峰度测试正态性。 该检验的零假设是，分布是从正态分布中得出的。"
        return ret

    '''
    检查两个分布是否相同
    '''
    @staticmethod
    def dist_test(ds1, ds2 , bins = 50, p_value = 0.05):
        ret = {}
        # F检验（对原始数据）
        ret['f-test'] = {}
        temp = f_oneway(ds1, ds2)
        ret['f-test']['is_same'] = str(temp.pvalue >= p_value)
        ret['f-test']['detail'] = temp
        ret['f-test']['describe'] = "F检验(方差分析)"

        # 卡方检验（先转换为直方图频数）
        hist1, _ = np.histogram(ds1, bins=bins)
        hist2, _ = np.histogram(ds2, bins=bins)

        # 为避免 sum 不一致，归一化 hist2 到与 hist1 相同的总和
        hist2_scaled = hist2 * (hist1.sum() / hist2.sum())

        ret['chis-test'] = {}
        temp = chisquare(f_obs=hist1, f_exp=hist2_scaled)
        ret['chis-test']['is_same'] = str(temp.pvalue >= p_value)
        ret['chis-test']['detail'] = temp
        ret['chis-test']['describe'] = f"卡方检验（将原始数据分为 {bins} 个箱后进行）"

        return ret
    '''
    正态拟合
    '''
    @staticmethod
    def normal_fit(ds_data,bins=100):
        data = pd.Series(ds_data).to_numpy()
        hist, bin_edges = np.histogram(data, bins=bins, density=True)
        bin_centers = 0.5*(bin_edges[1:] + bin_edges[:-1])
        
        def normal_pdf(x, mu, sigma):
            return norm.pdf(x, loc=mu, scale=sigma)
        def exponential_pdf(x, lamb):
            return expon.pdf(x, scale=1/lamb)
        
        params_normal, _ = curve_fit(normal_pdf, bin_centers, hist)
        params_exponential, _ = curve_fit(exponential_pdf, bin_centers, hist)
        
        plt.hist(data, bins=bins, density=True, alpha=0.5, color='blue')
        
        x_range = np.linspace(-4, 4, 1000)
        plt.plot(x_range, normal_pdf(x_range, *params_normal), color='red', label='Normal Fit')
        plt.plot(x_range, exponential_pdf(x_range, *params_exponential), color='green', label='Exponential Fit')
        
        plt.title('Fitting Distribution')
        plt.xlabel('Value')
        plt.ylabel('Density')
        plt.legend()
        
class CFADataTest:

    '''
    稳定性测试，平稳序列可以使用ARIMA
    1.判断是否为白噪声，白噪声满足ARIMA
        -- 平稳、独立和等方差
        -- 自相关和偏自相关函数在所有滞后阶数上接近于零
        -- 白噪声适合使用ARIMA模型进行预测
    2. 满足以下条件，可以使用ARIMA
        #ADF Test result同时小于1%、5%、10%
        #P-value (不变显著性) 接近0。
    '''
    @staticmethod
    def stationarity_test(timeSeries):
        dftest = adfuller(timeSeries)
        dfoutput = pd.Series(dftest[0:4], index=['Test Statistic','p-value','#Lags Used','Number of Observations Used'])
        for key,value in dftest[4].items():
            dfoutput['Critical Value (%s)'%key] = value
        del dfoutput['#Lags Used']
        del dfoutput['Number of Observations Used']
        return dfoutput

    '''
    绘制自相关ACF(与不同滞后阶数之间的相关性),和偏自相关性PACF（与特定滞后阶数的相关性）
    1. 自相关系数ACF：x轴表示滞后阶数（lag），y轴表示自相关系数
       -- 自相关性：自相关系数在某个滞后阶数上远离0，说明在这个滞后阶数上存在显著的自相关
       -- 显著性区域：自相关系数超过置信区间的边界，则认为该系数是显著的
       -- 截尾特性：随着滞后阶数的增加，自相关系数逐渐减小并趋于零，这种情况下，时间序列可能是平稳的，可以使用ARMA模型进行建模
       -- 周期性： 在某些滞后阶数上显示出周期性或周期性衰减。
    2. PACF
       -- 偏自相关性：x轴表示滞后阶数（lag），y轴表示偏自相关系数
       -- 显著性区域：如果偏自相关系数超过置信区间的边界，则认为该系数是显著的
       -- 截尾特性： 随着滞后阶数的增加，偏自相关系数逐渐减小并趋于零。这种情况下，时间序列可能是平稳的，可以使用AR模型进行建模
       -- 选择模型阶数：在某个滞后阶数上截尾，而在后续的滞后阶数上迅速衰减至接近零，那么该滞后阶数可能是适合的AR模型的阶数。
    '''
    @staticmethod
    def show_acf_pacf(timeSeries):
        plot_acf(timeSeries).show()
        plot_pacf(timeSeries).show()

    '''
    格兰特因果检验
    ds_result：结果序列（必须为平稳序列）
    ds_source：原因序列（必须为平稳序列）
    maxlag：时间间隔（整数或列表，为整数时，遍历所有的lag）
    返回值：
    1. 最小的p值
    2. 最佳lag
    3. approve_list通过测试的lag
    4. 详细测试结果
    '''
    @staticmethod
    def granger_test(ds_result,ds_source,maxlag,p_value = 0.05):

        df_test = pd.DataFrame()
        df_test['result'] = ds_result
        df_test['source'] = ds_source

        gc_result = grangercausalitytests(df_test[['result', 'source']], maxlag=maxlag,verbose=False )
        min_lag_p = 100
        best_lag = -1
        detail = {}
        approve_list = []
        for lag in gc_result:

            result = gc_result[lag][0]
            detail[lag] = result.copy()

            max_p = 0
            for test in result:
                test_value = result[test][0]
                test_p = result[test][1]
                if test_p > max_p:
                    max_p = test_p

            if max_p > p_value:
                continue
            #if max_p == 0:
            #    continue   
            approve_list.append({'lag':lag,"max p-value":max_p})
            
            if min_lag_p > max_p:
                min_lag_p = max_p
                best_lag = lag

        return min_lag_p, best_lag, approve_list, detail

class CFAFitter:
    '''
    对数据进行拟合
    fitter：拟合方式（linear，polynomial，exponential）
    degree：多项式阶数
    '''
    #linear,polynomial,exponential
    def __init__(self, fitter='polynomial',degree=2):
        self.params = None
        self.fitter = fitter
        self.degree = degree
        if fitter == 'linear':
            self.model_func = self.linear
        elif fitter == 'polynomial':
            self.model_func = self.polynomial
        elif fitter == 'exponential':
            self.model_func = self.exponential
        else:
            raise ValueError("Unsupported model type. Choose 'linear', 'polynomial', or 'exponential'.")

    @staticmethod
    def linear(x, a, b):
        return a * x + b

    @staticmethod
    def polynomial(x, *coeffs):
        return sum(c * x**i for i, c in enumerate(coeffs))

    @staticmethod
    def exponential(x, a, b):
        return a * np.exp(b * x)
    
    def get_fit_data(self,x_data,params):
        return self.model_func(x_data, *params) 

    def get_fit_param(self):
        return self.params

    def fit(self, x_data, y_data):
        if  self.fitter == 'linear':
            self.params, _ = curve_fit(self.linear, x_data, y_data)
            self.model_func = self.linear
        elif  self.fitter == 'polynomial':
            self.params, _ = curve_fit(self.polynomial, x_data, y_data, p0=[1] * (self.degree + 1))
            #self.params, _ = curve_fit(self.polynomial, self.x_data, self.y_data, p0=[1]*degree)
            self.model_func = self.polynomial
        elif  self.fitter == 'exponential':
            self.params, _ = curve_fit(self.exponential, x_data, y_data)
            self.model_func = self.exponential
        else:
            raise ValueError("Unsupported model type. Choose 'linear', 'polynomial', or 'exponential'.")
        return self.params

    def plot(self,x_data,y_data, params = None):
        plt.scatter(x_data, y_data, label='Data', color='blue', s=10)
        if params == None:
            params = self.get_fit_param()
        y_fit = self.get_fit_data(x_data,params)
        plt.plot(x_data, y_fit, label='Fitted Curve', color='red')
        plt.title(f'Fit: {self.model_func.__name__}')
        plt.xlabel('x')
        plt.ylabel('y')
        plt.legend()

class CFACommonStats:

    @staticmethod
    def get_one_stats(x, q=(0.25, 0.5, 0.75), autocorr_lags=(1,), add_corr=False):
        def corr_safe(df):
            num = df.select_dtypes(include="number").copy()
            num = num.dropna(axis=1, how="all")
            num = num.loc[:, num.nunique(dropna=True) > 1]
            if num.shape[1] < 2:
                return None

            with np.errstate(invalid="ignore", divide="ignore"):
                C = num.corr()
            return C

        def autocorr_safe(s: pd.Series, lag: int):
            s = pd.to_numeric(s, errors="coerce")
            s2 = s.dropna()
            if len(s2) <= lag + 1:
                return np.nan
            if s2.std(ddof=0) == 0:
                return np.nan
            with np.errstate(invalid="ignore", divide="ignore"):
                return s2.autocorr(lag=lag)

        def _one(s: pd.Series) -> pd.Series:
            s = s.copy()
            out = {}

            out["count"] = int(s.count())
            out["n"] = int(len(s))
            out["n_missing"] = int(s.isna().sum())
            out["missing_rate"] = (out["n_missing"] / out["n"]) if out["n"] else np.nan
            out["nunique"] = int(s.nunique(dropna=True))

            if not pd.api.types.is_numeric_dtype(s) and not pd.api.types.is_bool_dtype(s):
                m = s.mode(dropna=True)
                out["mode"] = m.iloc[0] if len(m) else np.nan
                return pd.Series(out)

            s_num = s.astype("float") if pd.api.types.is_bool_dtype(s) else pd.to_numeric(s, errors="coerce")

            out["mean"] = s_num.mean()
            out["median"] = s_num.median()
            m = s_num.mode(dropna=True)
            out["mode"] = m.iloc[0] if len(m) else np.nan

            out["min"] = s_num.min()
            out["max"] = s_num.max()
            out["range"] = (out["max"] - out["min"]) if pd.notna(out["max"]) and pd.notna(out["min"]) else np.nan

            qs = s_num.quantile(list(q))
            for qi, val in qs.items():
                out[f"q_{qi:g}"] = val
            out["iqr"] = (qs.loc[0.75] - qs.loc[0.25]) if (0.25 in qs.index and 0.75 in qs.index) else np.nan

            out["var"] = s_num.var()
            out["std"] = s_num.std()
            out["cv"] = (out["std"] / out["mean"]) if pd.notna(out["mean"]) and out["mean"] != 0 else np.nan

            out["mad_mean"] = (s_num - out["mean"]).abs().mean()
            out["skew"] = s_num.skew()
            out["kurt_fisher"] = s_num.kurt()
            out["kurt_pearson"] = out["kurt_fisher"] + 3 if pd.notna(out["kurt_fisher"]) else np.nan

            for lag in autocorr_lags:
                out[f"autocorr_lag{lag}"] = autocorr_safe(s_num, lag)

            return pd.Series(out)

        if isinstance(x, pd.Series):
            return _one(x).to_frame().T

        if isinstance(x, pd.DataFrame):
            stats = pd.DataFrame([_one(x[c]).rename(c) for c in x.columns])

            if add_corr:
                C = corr_safe(x)
                if C is not None:
                    corr_long = C.where(np.triu(np.ones(C.shape), 1).astype(bool)).stack()
                    corr_row = corr_long.rename(lambda idx: f"corr__{idx[0]}__{idx[1]}").to_frame().T
                    corr_row.index = ["__corr__"]
                    stats = pd.concat([stats, corr_row], axis=0, sort=False)

            return stats

        return get_one_stats(pd.Series(x), q=q, autocorr_lags=autocorr_lags)
    
    @staticmethod
    def get_stats(df_data):
        result = {}
        numeric_columns = df_data.select_dtypes(include=['number'])
        for key in numeric_columns:
            result[key] = {}
            t = CFACommonStats.get_one_stats(df_data[key]).to_dict(orient='records')[0]
            result[key]['mean'] = t['mean']
            result[key]['skew'] = t['skew']
            result[key]['kurt_fisher'] = t['kurt_fisher']
            result[key]['kurt_pearson'] = t['kurt_pearson']
            result[key]['autocorr_lag1'] = t['autocorr_lag1']
            result[key]['range'] = t['range']
            result[key]['min'] = t['min']
            result[key]['max'] = t['max']
            result[key]['iqr'] = t['iqr']
            result[key]['std'] = t['std']
            result[key]['cv'] = t['cv']
            
        return result

def main():
    #生成测试数据
    np.random.seed(11)
    X_Normal = np.random.randn(1000)
    X_linear = np.linspace(1,100,1000)
    df_data = pd.DataFrame()
    df_data['Normal'] = pd.Series(X_Normal)
    df_data['Linear'] = pd.Series(X_linear)

    #是否为高斯分布
    test = CFADataDistribution.normal_test(df_data['Normal'])
    print(json.dumps(test,indent=4))
    test = CFADataDistribution.dist_test(df_data['Normal'],df_data['Normal'])
    print(json.dumps(test,indent=4))
    CFADataDistribution.show_dist(df_data['Normal'],bins = 20)

    #平稳性检查
    df_test = CFADataTest.stationarity_test(df_data.Normal)
    print(df_test)
    CFADataTest.show_acf_pacf(df_data.Normal)
    plt.show()

    #数据拟合
    np.random.seed(0)
    x_data = np.linspace(0, 10, 100)
    y_data_linear = 2 * x_data + 1 + np.random.normal(size=x_data.size)
    y_data_polynomial = 1 * x_data**2 - 2 * x_data + 1 + np.random.normal(size=x_data.size)
    y_data_exponential = 2 * np.exp(0.5 * x_data) + np.random.normal(size=x_data.size)

    # 线性拟合
    linear_fitter = CFAFitter(fitter='linear')
    linear_params = linear_fitter.fit(x_data,y_data_linear)
    print("Linear fit parameters:", linear_params)
    linear_fitter.plot(x_data,y_data_linear)

    # 多项式拟合
    poly_fitter = CFAFitter(fitter='polynomial')
    poly_params = poly_fitter.fit(x_data,y_data_polynomial)
    print("Polynomial fit parameters:", poly_params)
    poly_fitter.plot(x_data,y_data_polynomial)

    # 指数拟合
    exp_fitter = CFAFitter(fitter="exponential")
    exp_params = exp_fitter.fit(x_data,y_data_exponential)
    print("Exponential fit parameters:", exp_params)
    exp_fitter.plot(x_data,y_data_exponential)
    plt.show()

    # 计算常见统计量
    stats = CFACommonStats.get_stats(df_data)
    print(stats)


if __name__ == "__main__":
    main()



'''
自动化训练分类模型（支持多分类），包括
1.模型训练
2.模型评估
3.ROC曲线绘制
'''
import time,json,os,sys
import numpy as np
import pandas as pd
import pandas as pd
import json,os,sys,datetime,warnings
from datetime import datetime
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from tqdm.notebook import tqdm
from sklearn.datasets import make_regression,make_classification
from sklearn.metrics import auc,recall_score,confusion_matrix,matthews_corrcoef,accuracy_score,precision_score,roc_auc_score,log_loss,f1_score,roc_curve,fbeta_score
from sklearn.metrics import confusion_matrix
import matplotlib.pyplot as plt
import seaborn as sns
sys.path.append(os.path.dirname(os.path.dirname(__file__)))
from FreeAeonML.FACommon import CFACommon
from FreeAeonML.FASample import CFASample

np.set_printoptions(suppress=True)
pd.set_option('display.float_format',lambda x : '%.8f' % x)
# 使用文泉驿字体
plt.rcParams['font.sans-serif'] = ['WenQuanYi Zen Hei']  
# 解决负号'-'显示为方块的问题
plt.rcParams['axes.unicode_minus'] = False

class CFASimilarity(object):
    def __init__(self,list1,list2):
        self.m_list1 = list1
        self.m_list2 = list2

    #余弦距离（cosine）
    def Cosine(self):
        ret = cosine_similarity([self.m_list1, self.m_list2])
        return ret[0][1]
        return cosine(self.m_list1, self.m_list2)

    #皮尔森相关系数（pearson）,https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.pearsonr.html
    def Pearson(self):
        return stats.pearsonr(self.m_list1, self.m_list2)[0]

    #欧式距离
    def Euclidean(self):
        return np.linalg.norm(np.array(self.m_list1) - np.array(self.m_list2))

    #KS检验,(P越大，两个分布越相似)
    #P比指定的显著水平（假设为5%）小，则我们完全可以拒绝假设，即两个分布不服从同一分布。
    #ret[0]越小越相似
    def KSTest(self):
        ret = ks_2samp(self.m_list1, self.m_list2)
        return ret[0],ret[1]

    #EMD距离
    def EMD(self):
        return wasserstein_distance(self.m_list1, self.m_list2)

    #Manhattan
    def Manhattan(self):
        return sum(abs(a-b) for a,b in zip(self.m_list1,self.m_list2))

    #Minkowski
    def Minkowski(self):
        return float(self.minkowski_distance(np.array(self.m_list1),np.array(self.m_list2),3))

    #Jaccard
    def Jaccard(self):
        return self.jaccard_similarity(self.m_list1,self.m_list2)

    def minkowski_distance(self,x,y,p_value):
        return self.nth_root(sum(pow(abs(a-b),p_value) for a,b in zip(x, y)),p_value)

    def nth_root(self,value, n_root):
        root_value = 1/float(n_root)
        return round (float(value) ** float(root_value),3)

    def jaccard_similarity(self,x,y):
        intersection_cardinality = len(set.intersection(*[set(x), set(y)]))
        union_cardinality = len(set.union(*[set(x), set(y)]))
        return intersection_cardinality/float(union_cardinality)

    #def Quadratic(self):
    #    return get_quadratic_distance(self.m_list1,self.m_list2)

class CFAEvaluation():
    
    def __init__(self):
        pass
    
    @staticmethod
    def evaluate(ds_pred,ds_true,average='weighted'):
        result = []
        y_pred = ds_pred.tolist()
        y_true = ds_true.tolist()
        tmp = {}
        tmp['confusion_matrix'] = confusion_matrix(y_true,y_pred)
        tmp['recall'] = recall_score(y_true,y_pred,average=average)
        tmp['mcc'] = matthews_corrcoef(y_true,y_pred)
        tmp['accuracy'] = accuracy_score(y_true,y_pred)
        tmp['precision'] = precision_score(y_true,y_pred,average=average)
        if ds_true.nunique() == 1:
            tmp['auc'] = "error true: 1 class"
            tmp['f1_score'] = "error True: 1 class"
            tmp['fbeta_score'] = "error True: 1 class"
        elif ds_pred.nunique() == 1:
            tmp['auc'] = "error pred: 1 class"
            tmp['f1_score'] = "error pred: 1 class"
            tmp['fbeta_score'] = "error pred: 1 class"
        elif ds_true.nunique() > 2:
            tmp['auc_ovo'] = roc_auc_score(y_true,y_prab,multi_class='ovo',average=average)
            tmp['auc_ovr'] = roc_auc_score(y_true,y_prab,multi_class='ovr',average=average)
        else:
            tmp['auc'] = roc_auc_score(y_true,y_pred,average=average)
            tmp['f1_score'] = f1_score(y_true,y_pred,average=average)
            tmp['fbeta_score'] = fbeta_score(y_true,y_pred,beta=0.5,average=average)
        result.append(tmp)
        return pd.DataFrame(result)
        
    @staticmethod
    def get_binary_roc(ds_true,ds_prob):
        fpr, tpr, thresholds = roc_curve(ds_true, ds_prob)
        roc_auc = auc(fpr, tpr)
        df_roc = pd.DataFrame()
        df_roc['fpr'] = fpr
        df_roc['tpr'] = tpr
        df_roc['thresholds'] = thresholds
        return roc_auc,df_roc
    
    @staticmethod
    def show_binary_roc(roc_auc,df_roc,title = None):
        plt.figure()
        plt.plot(df_roc['fpr'], df_roc['tpr'], color='blue', label='ROC curve (area = {:.2f})'.format(roc_auc))
        plt.plot([0, 1], [0, 1], color='red', linestyle='--')  # 对角线
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.0])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        if title != None:
            plt.title('ROC of %s'%title)
        else:
            plt.title('Receiver Operating Characteristic (ROC) Curve')
        plt.legend(loc='lower right')
        plt.grid()
        plt.show()

def main():
    df_sample = CFASample.get_random_classification(1000,n_feature=10,n_class=2)
    df_train,df_test = CFASample.split_dataset(df_sample)

if __name__ == "__main__":
    main()



'''
自动化训练分类模型
1.模型训练
2.模型评估
'''
import platform
import numpy as np
import pandas as pd
import pickle,h2o,json,os,sys,time,warnings
import matplotlib.pyplot as plt
import seaborn as sns
from tqdm import tqdm
from h2o.automl import H2OAutoML
from h2o.estimators import *
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score
import matplotlib.pyplot as plt
sys.path.append(os.path.dirname(os.path.dirname(__file__)))
from FreeAeonML.FACommon import CFACommon
from FreeAeonML.FASample import CFASample
np.set_printoptions(suppress=True)
pd.set_option('display.float_format',lambda x : '%.8f' % x)

'''
使用h2o进行automl
'''
class CFAModelRegression():
  
    def __init__(self,ip='localhost',port=54321,models = None):
        
        #h2o.init(nthreads = -1, verbose=False)
        #h2o.connect(ip=ip,port=port,verbose=False)
        if models == None:
            self.m_models = {
                "rf":H2ORandomForestEstimator(),
                "ann":H2ODeepLearningEstimator(),
                "glm":H2OGeneralizedLinearEstimator(),
                "gbm":H2OGradientBoostingEstimator()
                #"xgboost":H2OXGBoostEstimator()
            }
            # 判断平台与可用性
            system = platform.system()
            machine = platform.machine()

            try:
                if H2OXGBoostEstimator.available():
                    if system != "Windows" and not machine.startswith("arm"):
                        self.m_models["xgboost"] = H2OXGBoostEstimator()
                    else:
                        print("⚠️ 当前平台不支持 H2O XGBoost（Windows 或 ARM 架构）")
                else:
                    print("⚠️ H2O XGBoostEstimator 不可用（编译未启用或未安装）")
            except Exception as e:
                print(f"⚠️ 检查 H2O XGBoost 可用性失败：{e}")

        else:
            self.m_models = models

    def train(self,df_sample,x_columns = [] ,y_column='y',train_ratio = 0.85):
        
        if x_columns == []:
            total_columns = df_sample.keys().tolist()
        else:
            total_columns = x_columns
            if not y_column in total_columns:
                total_columns.append(y_column)
            
        df_temp = df_sample[total_columns]
        
        df_h2o = h2o.H2OFrame(df_temp)
        if train_ratio > 0:
            df_train, df_valid = df_h2o.split_frame(ratios=[train_ratio], seed=1234)
        else:
            df_train = df_h2o
            
        x = df_train.columns
        y = y_column
        x.remove(y)
        
        for key in self.m_models:
            model = self.m_models[key]
            #print("begin train ",key)
            if train_ratio > 0:
                model.train(x=x, y=y,training_frame=df_train,validation_frame=df_valid)
            else:
                model.train(x=x, y=y,training_frame=df_train)
            #print("end train ",key)
    
    def load(self,model_path):
        for key in self.m_models:
            folder = "%s/%s"%(os.path.abspath(model_path),key)
            model = self.m_models[key]
            all_models = os.listdir(folder)
            all_models.sort()
            for file_name in all_models:
                model_file = os.path.join(folder, file_name)
                print("loading",model_file)
                self.m_models[key] = h2o.load_model(model_file)
   
    def save(self,model_path):
        for key in self.m_models:
            folder = "%s/%s"%(os.path.abspath(model_path),key)
            model = self.m_models[key]
            if not os.path.exists(folder):
                os.makedirs(folder)
                
            for file_name in os.listdir(folder):
                model_file = os.path.join(folder, file_name)
                os.remove(model_file)
            
            model_saved = h2o.save_model(model=model, path=folder, force=True)
            print("save model to ",model_saved)

    def predict(self,df_sample,x_columns = [],y_column='y'):
        
        if x_columns == []:
            total_columns = df_sample.keys().tolist()
        else:
            total_columns = x_columns
         
        if y_column in total_columns:
            total_columns.remove(y_column)
                
        if y_column in df_sample.keys().tolist():
            y_true = df_sample[y_column].tolist()
        else:
            y_true = None
            
        df_temp = df_sample[total_columns]
        df_h2o = h2o.H2OFrame(df_temp)

        result = []
        for key in self.m_models:
            model = self.m_models[key]
            training_columns = model._model_json['output']['names'][:-1]
            for col in training_columns:
                if col not in df_temp.columns:
                    print("NO training column %s found not"%col)
                    df_temp = df_temp.reindex(columns=training_columns, fill_value=0)  
                    df_h2o = h2o.H2OFrame(df_temp)
                    break
            df_t = model.predict(df_h2o).as_data_frame()
            df_t['model'] = key
            df_t['true'] = y_true
            result.extend(json.loads(df_t.to_json(orient='records')))
            
        return pd.DataFrame(result)
    
    def evaluate(self,df_sample,x_columns = [],y_column='y'):
        df_pred = self.predict( df_sample,x_columns=x_columns , y_column=y_column )
        result = []
        for model,df in df_pred.groupby('model'):
            y_pred = df['predict'].tolist()
            y_true = df['true'].tolist()
            tmp = {}
            tmp['model'] = model
            tmp['mse'] = mean_squared_error(y_true,y_pred)
            tmp['rmse'] = np.sqrt(tmp['mse'])
            tmp['mae'] = mean_absolute_error(y_true,y_pred)
            tmp['r2'] = r2_score(y_true,y_pred)
            result.append(tmp)

        return pd.DataFrame( result )

    def importance(self):
        result = []
        for key in self.m_models:
            model = self.m_models[key]
            df_temp = model.varimp(use_pandas=True)
            if type(df_temp) == pd.DataFrame:
                df_temp['model'] = key
                result.extend(json.loads(df_temp.to_json(orient='records')))
        df_importance = pd.DataFrame(result).reset_index(drop=True)
        return df_importance

def main():
    
    #如果是WSL,注释掉h2o.init(),使用h2o.connect()
    h2o.init(nthreads = -1, verbose=False)
    #h2o.connect(ip=ip,port=port)

    df_sample = CFASample.get_random_regression(1000)
    df_sample['y'] = df_sample['y'].astype(float)
    df_train,df_test = CFASample.split_dataset(df_sample)
    print(df_train)

    model_1 = CFAModelRegression()
    model_1.train(df_train,y_column='y')

    df_pred = model_1.predict(df_sample,y_column='y')
    print(df_pred)

    df_verify = model_1.evaluate(df_sample,y_column='y')
    print(df_verify)

    df_importance = model_1.importance()
    print(df_importance)

if __name__ == "__main__":
    main()



from setuptools import setup, find_packages
import os

def parse_requirements(filename):
    with open(filename, 'r') as file:
        return [line.strip() for line in file if line and not line.startswith('#')]

setup(
    name="FreeAeon-ML",
    version="0.3.4",
    author="Jim Xie",
    author_email="jim.xie.cn@outlook.com",
    description="A comprehensive machine learning toolkit for data analysis, preprocessing, modeling, and evaluation.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/jim-xie-cn/FreeAeon-ML",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=parse_requirements('requirements.txt'),
    include_package_data=True,
    package_data={
        "": ["quick_start.py","images/*","tests/*","FreeAeonML/*"],
    },
    license='MIT'
)



'''
ARIMA处理类
'''
import matplotlib.pyplot as plt
import pandas as pd
import statsmodels.api as sm
from statsmodels.tsa.seasonal import seasonal_decompose
import numpy as np
from tqdm import tqdm
from itertools import product
import os,sys
sys.path.append(os.path.dirname(os.path.dirname(__file__)))
from FreeAeonML.FADataPreprocess import CFADataPreprocess

class CFAArima:
    
    def __init__(self):
        pass
        
    @staticmethod
    def fit(dataseries,p=1,d=0,q=1):
        train_data = dataseries
        train_data = train_data.astype("float64")
        return sm.tsa.arima.ARIMA(train_data, order=(p, d, q)).fit()

    '''
    显示模型信息，关注以下指标
    AIC，越大越好
    BIC，越大越好
    P>|z|，越小越好
    Prob(Q)
    '''
    @staticmethod
    def show_model(result_arima):
        print(result_arima.summary())
    
    @staticmethod
    def predict(result_arima):
        return result_arima.predict()
        
    @staticmethod
    def forecast(result_arima,num_future_points=1):
        return result_arima.forecast(steps=num_future_points)
        
    @staticmethod
    def show_result(train_data,predicted):
        plt.figure(facecolor='white')
        error = (np.sqrt(sum((predicted-train_data).dropna()**2/train_data.size)))
        predicted.plot(color='blue', label='Predict')
        train_data.plot(color='red', label='Original')
        plt.legend(loc='best')
        plt.title('RMSE: %.4f'% error)
        plt.show()
    
    @staticmethod
    def auto_fit(ds_train,
                 ds_test,
                 seasonal_order_range=((1,2),(1,2),(1,10),(1,7))):
        ds_train = ds_train.astype("float64")
        ds_test = ds_test.astype("float64")
        order_p = seasonal_order_range[0]
        order_d = seasonal_order_range[1]
        order_q = seasonal_order_range[2]
        order_period = seasonal_order_range[3]
        all_combinations = list(product(order_p, order_d, order_q, order_period))
    
        ret_order = None
        min_error = np.inf
        for p,d,q,period in tqdm(all_combinations):
            try:
                model = sm.tsa.arima.ARIMA(ds_train, seasonal_order=(p, d, q,period))
                res_arima = model.fit()
                predicted = res_arima.forecast(ds_test.shape[0])
                error = (np.sqrt(sum((predicted-ds_test).dropna()**2/ds_test.size)))
                if error < min_error:
                    result_arima = res_arima
                    min_predicted = predicted
                    min_error = error
                    ret_order = (p,d,q,period)
            except:
                print("error",p,d,q,period)
                        
        return result_arima,ret_order

    @staticmethod
    def decomposition(ds_data,model='additive',period=7):
        decomposition = seasonal_decompose(ds_data,period=period,model=model)
        return decomposition

def main():
    ds_data = pd.Series([1,2,3,4,5,6])
    transformed_data = CFADataPreprocess.get_abnormal(ds_data)
    ds_data = ds_data.drop(transformed_data.index)
    ds_data = ds_data[ds_data>0]
    seasonal_order_range = ((1,4),(1,4),(3,10),(5,10))
    model,ret_order = CFAArima.auto_fit(ds_train = ds_data.head(4),
                                        ds_test = ds_data.tail(2),
                                        seasonal_order_range = seasonal_order_range)

    CFAArima.show_model(model)
    predict_data = model.predict()
    CFAArima.show_result(ds_data,predict_data)
    decomposition = CFAArima.decomposition(ds_data,period=2)
    decomposition.plot()
    plt.show()

if __name__ == "__main__":         
    main()     
